/* This code may be redistributed and reused; however, all credit
 * must be attributed to the original authors.
 * Authors: Craig Lombardo & Rachel Szteinberg '17
 * Date Created: 12/13/2016
 */

#include "config.h"
#include "tft_master.h"
#include "tft_gfx.h"
#include <stdlib.h>
#include "adc_intf.h"

// threading library
// config.h sets 40 MHz
#define	SYS_FREQ 40000000
#include "pt_cornell_1_2.h"

#include <time.h>
#include <stdio.h>

// BLE library
#include "biscuit_central.h"
#include "ble_hci.h"

#define NOPUSH 0
#define MAYBEPUSH 1
#define PUSHED 2
#define MAYBENOPUSH 3

//enum type for password recording states 

enum recordingState {
    IDLE,
    RESET_PARAMS,
    RECORDING,
    SEND,
    HOLD_DATA
};
//Default state is IDLE
enum recordingState r_state = IDLE;

//Action codes are the codes that are sent over Bluetooth for the lock to read in
//and interpret

enum ActionCodes {
    OFF,
    SINGLE,
    DOUBLE,
    HOLD,
    NORTH,
    SOUTH,
    EAST,
    WEST
};
//By default we set it to be OFF
enum ActionCodes actionCode = OFF;

//These are the prediction values
int x, y, z;

int x1 = 0;
int x2 = 0;
int x3 = 0;

//Used to determine whether a value is above, below or within a threshold
float x11;
float x21;
float x31;

//The initial x,y,z value used for reference
int xInit;
int yInit;
int zInit;

//Used to determine whether or not there was a prediction in a certain direction
//These variables are independent of one another
int hPred = 0;
int vPred = 0;

//These values hold the actual prediction values after all comparison logic
int hDir = 0;
int vDir = 0;

int canSendPress = 1;

//These are used to set an upper bound on the number of samples taken 
//while pattern matching
int maxSamples = 2000;
int sampleCount = 0;

//After these values we are reasonably certain we have gotten a match
int predictionThresh = 60;

//These are used to help determine skew correction and direction predictions
int xThresh = 15;
int yThresh = 1000;
int zThresh = 15;

int pushState = 0;
int acted = 0;

//Double click timeout
int timeout = 300;

int singlePress = 0;
int doublePress = 0;
int buttonHeld = 0;
int buttonPressed = 0;

//Global i for any and all for loops
int i = 0;

// definitions for connection handles
#define KEY 0xFFFE
#define LOCK 0x0000
uint8_t lock_address[6] = {0x23, 0x11, 0x7B, 0x4C, 0x99, 0xB4};
uint8_t writeBuf[4];

// thread control structs
static struct pt pt_read_accel, pt_user_input, pt_push_button;

//Authors: Brandon Smith and Nick Falco
//delays during init

void delay(unsigned t) { //uses internal timer 2
    T2CON = 0x8000; // enable tmr1, Tcy, 1:1
    while (t--) {
        TMR2 = 0;
        // while (TMR1<8000);
        while (TMR2 < 8000);
    }
}

/* This method will return a value corresponding to whether or not a value
 * is between, above or below a certain threshold +- some offset
 * Authors: Craig Lombardo & Rachel Szteinberg
 */
int checkVal(float val, int middle, int offset) {
    if (val > (middle + offset)) return 1;
    else if (val < (middle - offset)) return 2;
    else return 0;
}

/* This method will change the color of the LED based on the binary representation
 * of the number input where:
 * RA3 = Red
 * RB4 = Green
 * RA4 = Blue
 * Authors: Craig Lombardo & Rachel Szteinberg
 */
static void changeLED(int led) {
    //Red = 4
    if (led >> 2 & 0x1) mPORTASetBits(BIT_3);
    else mPORTAClearBits(BIT_3);

    //Green = 2
    if (led >> 1 & 0x1) mPORTBSetBits(BIT_4);
    else mPORTBClearBits(BIT_4);

    //Blue = 1
    if (led & 0x1) mPORTASetBits(BIT_4);
    else mPORTAClearBits(BIT_4);
}

/* This method will determine if the debounced button has been pressed.
 * Authors: Craig Lombardo & Rachel Szteinberg
 */
static void checkDebouncedPress() {
    switch (pushState) {
        case NOPUSH:
            if (buttonPressed) pushState = MAYBEPUSH;
            else pushState = NOPUSH;
            break;
        case MAYBEPUSH:
            if (buttonPressed) {
                //Debounced actions go here
                //First time we set single press high
                //Second time double press high
                pushState = PUSHED;
                if (singlePress) doublePress = 1;
                else singlePress = 1;
            } else {
                pushState = NOPUSH;
                canSendPress = 1;
            }
            break;
        case PUSHED:
            if (buttonPressed) pushState = PUSHED;
            else pushState = MAYBENOPUSH;
            break;
        case MAYBENOPUSH:
            if (buttonPressed) pushState = PUSHED;
            else {
                //Button release actions go here
                pushState = NOPUSH;
                buttonHeld = 0;
                actionCode = OFF;
                canSendPress = 1;

                singlePress = 0;
                doublePress = 0;
            }
            break;
    }
}

/* This thread will determine whether the button has been pressed once, twice, or
 * if it is being held currently.
 * Authors: Craig Lombardo & Rachel Szteinberg
 */
static PT_THREAD(protothread_push_button(struct pt *pt)) {
    PT_BEGIN(pt);
    while (1) {
        PT_YIELD_TIME_msec(10);
        if (canSendPress) {
            if (singlePress) {
                acted = 0;
                //if we get a second click in the timeout window we do not want
                //to wait
                for (i = 0; i < timeout && acted == 0; i++) {
                    PT_YIELD_TIME_msec(1);
                    if (doublePress) {
                        acted = 1;
                        actionCode = DOUBLE;
                        //Send the action code
                        sprintf(writeBuf, " %i ", actionCode);
                        biscuit_central_write_bytes(LOCK, writeBuf, sizeof (writeBuf));
                    }
                }
                //If a double click has not been registered by this point we know
                //they either single or double clicked
                if (!acted) {
                    if (buttonPressed) {
                        buttonHeld = 1;
                        actionCode = HOLD;
                    } else {
                        actionCode = SINGLE;
                    }
                    //Send the action code
                    sprintf(writeBuf, " %i ", actionCode);
                    biscuit_central_write_bytes(LOCK, writeBuf, sizeof (writeBuf));
                    canSendPress = 0;
                }

            }
        }
        PT_END(pt);
    }
}

/* This method checks debounced button and accelerometer data to see if button has been 
 * pressed, and if so which type of press, and if the accelerometer has 
 * detected. This is determining which action code will be sent to the lock. 
 * Authors: Craig Lombardo & Rachel Szteinberg
 */
static PT_THREAD(protothread_user_input(struct pt *pt)) {
    PT_BEGIN(pt);
    while (1) {
        PT_YIELD_TIME_msec(10);
        buttonPressed = mPORTAReadBits(BIT_0) ? 0 : 1;
        checkDebouncedPress();
    }
    PT_END(pt);
}

/* This method will read in accelerometer data and determine key's
 * direction of motion.
 * Authors: Craig Lombardo & Rachel Szteinberg
 */
static PT_THREAD(protothread_read_accel(struct pt *pt)) {
    PT_BEGIN(pt);

    while (1) {
        PT_YIELD_TIME_msec(2);

        //read accelerometer direction pins
        // AN 11 = PIN 24 = RB13 = X
        // AN 10 = PIN 25 = RB14 = Y
        // AN 9  = PIN 26 = RB15 = Z
        x11 = ((float) analogRead(11));
        x21 = ((float) analogRead(10));
        x31 = ((float) analogRead(9));

        //if the button is pressed (ie activate recording) save motion values
        if (buttonPressed) {
            x1 = checkVal(x11, xInit, xThresh); //x
            x2 = checkVal(x21, yInit, yThresh); //y
            x3 = checkVal(x31, zInit, zThresh); //z
        }
        //Recording state
        switch (r_state) {
            case IDLE:
                if (buttonPressed) {
                    //if button is pressed, initialize starting direction
                    xInit = ((float) analogRead(11));
                    yInit = ((float) analogRead(10));
                    zInit = ((float) analogRead(9));
                    r_state = RESET_PARAMS;
                } else {
                    //blue is idle
                    changeLED(1);
                    r_state = IDLE;
                }
                break;
            case RESET_PARAMS:
                changeLED(3);
                x = 0;
                y = 0;
                z = 0;
                hDir = 0;
                vDir = 0;
                sampleCount = 0;
                //We don't want to begin recording until we detect a deliberate motion
                //this helps with skew error prevention
                if ((checkVal(x11, xInit, xThresh / 4) != 0) || (checkVal(x31, zInit, zThresh / 4) != 0)) {
                    r_state = RECORDING;
                } else {
                    xInit = x11;
                    yInit = x21;
                    zInit = x31;
                    r_state = RESET_PARAMS;
                }
                break;
            case RECORDING:
                changeLED(5);
                //increment x, y, z values based on direction of motion
                x += ((x1 == 1) ? -1 : ((x1 == 2) ? 1 : 0));
                y += ((x2 == 1) ? 1 : ((x2 == 2) ? -1 : 0));
                z += ((x3 == 1) ? 1 : ((x3 == 2) ? -1 : 0));
                sampleCount++;
                r_state = buttonPressed && sampleCount <= maxSamples && abs(x) <= predictionThresh && abs(z) <= predictionThresh ? RECORDING : SEND;
                break;
            case SEND:
                //to send data, determine direction of change of motion
                if (x >= predictionThresh) hPred = 1;
                else if (x <= -predictionThresh) hPred = 2;
                else hPred = 0;

                if (z >= predictionThresh) vPred = 1;
                else if (z <= -predictionThresh) vPred = 2;
                else vPred = 0;

                //We use a value of 30 as a measure of being reasonably certain
                //that we have seen the intended direction
                if (abs(z) - abs(x) > 30) {
                    hDir = 0;
                    vDir = vPred;
                } else if (abs(x) - abs(z) > 30) {
                    hDir = hPred;
                    vDir = 0;
                } else {
                    hDir = 0;
                    vDir = 0;
                }

                if (!(hDir == 0 && vDir == 0)) {
                    if (hDir == 1) actionCode = EAST;
                    else if (hDir == 2) actionCode = WEST;
                    else if (vDir == 1) actionCode = NORTH;
                    else if (vDir == 2) actionCode = SOUTH;
                    sprintf(writeBuf, " %i ", actionCode);
                    biscuit_central_write_bytes(LOCK, writeBuf, sizeof (writeBuf));
                }
                //Don't want to stay in this stage if the user is holding the button
                r_state = buttonPressed ? HOLD_DATA : IDLE;
                break;
            case HOLD_DATA:
                //Want to stay in this stage if the user is holding the button
                r_state = buttonPressed ? HOLD_DATA : IDLE;
                break;
            default:
                r_state = IDLE;
        }
    }
    PT_END(pt);
}


/* This main method is used for the initial setup of the device.
 * Authors: Craig Lombardo & Rachel Szteinberg
 */
void main(void) {
    SYSTEMConfigPerformance(PBCLK);

    ANSELA = 0;
    ANSELB = 0;
    CM1CON = 0;
    CM2CON = 0;
    
    configureADC();

    mPORTBSetPinsDigitalIn(BIT_13 | BIT_14 | BIT_15);

    //RA3 = Red LED
    //RA4 = Blue LED
    //RB4 = Green LED
    mPORTASetPinsDigitalOut(BIT_3 | BIT_4);
    mPORTBSetPinsDigitalOut(BIT_4);

    mPORTAClearBits(BIT_3 | BIT_4);
    mPORTBClearBits(BIT_4);

    //RB7 = reset pin for BLE chip
    mPORTBSetPinsDigitalOut(BIT_7);

    //power LED status before connected = red
    changeLED(4);

    // === config threads ==========
    // turns OFF UART support and debugger pin
    PT_setup();

    // === setup system wide interrupts  ========
    INTEnableSystemMultiVectoredInt();

    // init the threads
    PT_INIT(&pt_read_accel);
    PT_INIT(&pt_user_input);
    PT_INIT(&pt_push_button);

    //reset BLE chip
    mPORTBSetBits(BIT_7);
    delay(2500);
    mPORTBClearBits(BIT_7);
    delay(2500);

    //set up BLE chip
    biscuit_central_init();
    delay(5000);

    biscuit_central_connect(lock_address);
    delay(40000);
    biscuit_central_enable_notification(LOCK);
    delay(2500);

    //white LED once connected
    changeLED(7);

    // round-robin scheduler for threads
    while (1) {
        PT_SCHEDULE(protothread_read_accel(&pt_read_accel));
        PT_SCHEDULE(protothread_user_input(&pt_user_input));
        PT_SCHEDULE(protothread_push_button(&pt_push_button));
    }
}